//! This module contains [`AtemPacket`] which is a zero*ish* copy abstraction over a [`AsRef`]`<[u8]>`.
use core::{fmt::Display, str};
use enumflags2::{BitFlags, bitflags};

use super::atem_field::{Field, FieldParsingError, RawField};

/// The "hello" packet to start communication with the ATEM
pub const COMMAND_CONNECT_HELLO: [u8; 20] = [
    0x10, 0x14, 0x53, 0xab, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3a, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
];

/// Maximum atem packet length. This is determined by the maximum value that can be stored
/// in the length field.
pub const MAX_LEN: usize = field::LEN_MASK as usize;

mod field {
    use core::ops::{Range, RangeInclusive};

    pub(crate) const FLAGS_LEN_H: usize = 0;
    pub(crate) const FLAGS_MASK: u8 = 0b1111_1000;
    pub(crate) const LEN_L: usize = 1;
    pub(crate) const LEN: RangeInclusive<usize> = FLAGS_LEN_H..=LEN_L;
    pub(crate) const LEN_MASK: u16 = 0x7ff;

    pub(crate) const SESSION_ID: Range<usize> = 2..4;
    pub(crate) const ACK_NUMBER: Range<usize> = 4..6;
    pub(crate) const REMOTE_SEQ_NUM: Range<usize> = 8..10;
    pub(crate) const LOCAL_SEQ_NUM: Range<usize> = 10..12;
}

/// An ATEM protocol packet
#[derive(Debug)]
pub struct AtemPacket<T: AsRef<[u8]>> {
    buf: T,
}

impl<'a> TryFrom<&'a [u8]> for AtemPacket<&'a [u8]> {
    type Error = AtemPacketParseError;
    fn try_from(buf: &'a [u8]) -> Result<Self, Self::Error> {
        AtemPacket::new_checked(buf)
    }
}

#[derive(Debug)]
pub enum AtemPacketParseError {
    /// The packet was too short
    TooShort {
        got: usize,
    },
    /// The packet's stated and actual lengths were different
    LengthDiffers {
        expected: u16,
        got: usize,
    },
    InvalidFlags,
}

/// ATEM packet flags
#[bitflags]
#[repr(u8)]
#[derive(PartialEq, Copy, Clone, Debug)]
pub enum PacketFlag {
    /// AKA "reliable". This packet should be ACKed by the other end
    AckRequest = 0x1,
    /// AKA "syn" - used in the connection handshake
    NewSessionId = 0x2,
    /// This packet is a retransmit of a previous one
    IsRetransmit = 0x4,
    /// Request retransmission of a previous sequence ID
    RetransmitRequest = 0x8,
    /// This packet is an ACK
    AckReply = 0x10,
}

/// An error while constructing a packet
#[derive(Debug)]
pub enum AtemPacketInitError {
    /// The provided buf was too small
    BufTooSmall { need: usize, got: usize },
    /// The provided payload was too long
    DataTooLong { got: usize, max: usize },
}

impl<'a> AtemPacket<&'a mut [u8]> {
    /// Initialise an [`AtemPacket`] into the given `buf`, with optional payload copied from `data`.
    /// Returns the constructed packet, and the remaining spare space in `buf`, if there was any.
    pub fn init<'b>(
        buf: &'a mut [u8],
        flags: BitFlags<PacketFlag>,
        session_id: u16,
        local_seq_num: u16,
        data: Option<&'b [u8]>,
    ) -> Result<(Self, &'a mut [u8]), AtemPacketInitError> {
        let len = 12 + data.map_or(0, |d| d.len());
        if len > buf.len() {
            return Err(AtemPacketInitError::BufTooSmall {
                need: len,
                got: buf.len(),
            });
        }
        if len > MAX_LEN {
            return Err(AtemPacketInitError::DataTooLong {
                got: len - 12,
                max: MAX_LEN - 12,
            });
        }
        let (pkt, rem) = buf.split_at_mut(len);
        let mut p = AtemPacket { buf: pkt };
        p.set_flags(flags)
            .set_session_id(session_id)
            .set_local_seq_num(local_seq_num)
            .set_len(len.try_into().unwrap());
        if let Some(d) = data {
            p.set_data(d);
        }
        Ok((p, rem))
    }
}

impl<T: AsRef<[u8]> + AsMut<[u8]>> AtemPacket<T> {
    pub fn set_flags(&mut self, flags: BitFlags<PacketFlag>) -> &mut Self {
        let prev = self.buf.as_ref()[field::FLAGS_LEN_H];
        self.buf.as_mut()[field::FLAGS_LEN_H] = (prev & !field::FLAGS_MASK) | (flags.bits() << 3);
        self
    }

    pub fn set_len(&mut self, value: u16) -> &mut Self {
        let v = value.to_be_bytes();
        self.buf.as_mut()[field::FLAGS_LEN_H] &= v[0] | field::FLAGS_MASK;
        self.buf.as_mut()[field::LEN_L] = v[1];
        self
    }

    pub fn set_session_id(&mut self, value: u16) -> &mut Self {
        self.buf.as_mut()[field::SESSION_ID].copy_from_slice(&value.to_be_bytes());
        self
    }
    pub fn set_local_seq_num(&mut self, value: u16) -> &mut Self {
        self.buf.as_mut()[field::LOCAL_SEQ_NUM].copy_from_slice(&value.to_be_bytes());
        self
    }
    pub fn set_remote_seq_num(&mut self, value: u16) -> &mut Self {
        self.buf.as_mut()[field::REMOTE_SEQ_NUM].copy_from_slice(&value.to_be_bytes());
        self
    }
    pub fn set_data<'a>(&mut self, value: &'a [u8]) -> &mut Self {
        self.buf.as_mut()[12..].copy_from_slice(value);
        self
    }

    pub fn set_ack_num(&mut self, value: u16) -> &mut Self {
        self.buf.as_mut()[field::ACK_NUMBER].copy_from_slice(&value.to_be_bytes());
        self
    }
}

impl<T: AsRef<[u8]>> AtemPacket<T> {
    pub fn new_checked(buf: T) -> Result<Self, AtemPacketParseError> {
        let len = buf.as_ref().len();
        if len < 12 {
            return Err(AtemPacketParseError::TooShort {
                got: buf.as_ref().len(),
            });
        }
        let p = Self { buf };
        if p.length() as usize != len {
            return Err(AtemPacketParseError::LengthDiffers {
                expected: p.length(),
                got: len,
            });
        }
        // Check flags are valid
        let _: BitFlags<PacketFlag> = p
            .flags_raw()
            .try_into()
            .map_err(|_| AtemPacketParseError::InvalidFlags)?;
        Ok(p)
    }

    /// Consumes self, returning the inner packet buffer
    pub fn inner(self) -> T {
        self.buf
    }
    pub fn length(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[field::LEN].try_into().unwrap()) & field::LEN_MASK
    }

    fn flags_raw(&self) -> u8 {
        self.buf.as_ref()[field::FLAGS_LEN_H] >> 3
    }

    pub fn flags(&self) -> BitFlags<PacketFlag> {
        // We `unwrap` here, but given we check the flags in the constructor this should never panic.
        self.flags_raw().try_into().unwrap()
    }

    pub fn session_id(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[field::SESSION_ID].try_into().unwrap())
    }

    pub fn ack_number(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[field::ACK_NUMBER].try_into().unwrap())
    }

    pub fn remote_sequence_number(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[field::REMOTE_SEQ_NUM].try_into().unwrap())
    }

    pub fn local_sequence_number(&self) -> u16 {
        u16::from_be_bytes(self.buf.as_ref()[field::LOCAL_SEQ_NUM].try_into().unwrap())
    }

    pub fn retransmit_request(&self) -> Option<u16> {
        self.flags()
            .contains(PacketFlag::RetransmitRequest)
            .then_some(u16::from_be_bytes([
                self.buf.as_ref()[6],
                self.buf.as_ref()[7],
            ]))
    }

    pub fn ack_reply(&self) -> Option<u16> {
        self.flags()
            .contains(PacketFlag::AckReply)
            .then_some(self.ack_number())
    }

    /// Get an iterator over the `Field`s in this packet.
    ///
    /// Returns None if this is a packet without fields.
    pub fn raw_fields(&self) -> RawFields {
        // TODO: do we only ever get newsessionid during the handshake (i.e. not in a packet with fields)?
        let has_fields = !self.flags().contains(PacketFlag::NewSessionId);
        RawFields::new(if has_fields { self.body() } else { &[] })
    }

    pub fn fields<'a, F: Field<'a>>(
        &'a self,
    ) -> impl Iterator<Item = Result<F, FieldParsingError>> + use<'a, F, T> {
        self.raw_fields().filter_map(|f| match F::try_from_raw(f) {
            Err(FieldParsingError::MismatchedFieldType) => None,
            x => Some(x),
        })
    }

    pub fn body(&self) -> &[u8] {
        &self.buf.as_ref()[12..]
    }
}

impl<T: AsRef<[u8]>> Display for AtemPacket<T> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(
            f,
            "len: {}, sid: {}, flags: {}, ack: {}, localseq: {}, remoteseq: {}",
            self.length(),
            self.session_id(),
            self.flags(),
            self.ack_number(),
            self.remote_sequence_number(),
            self.local_sequence_number(),
        )
    }
}

/// Created by [`AtemPacket::raw_fields`]
pub struct RawFields<'a> {
    data: &'a [u8],
    // The offset of the next field in the packet
    offset: usize,
}

impl<'a> RawFields<'a> {
    pub(crate) fn new(data: &'a [u8]) -> Self {
        Self { data, offset: 0 }
    }
}

impl<'a> Iterator for RawFields<'a> {
    type Item = RawField<'a>;
    fn next(&mut self) -> Option<Self::Item> {
        let remain = self.data.len() - self.offset;
        if remain == 0 {
            return None;
        } else if remain < 8 {
            // TODO: is 8 indeed the minimum size for something here? (i.e. no field data)
            panic!("Oh no");
        }

        let length =
            u16::from_be_bytes(self.data[self.offset..=self.offset + 1].try_into().unwrap());
        // TODO: sanity check length
        let r#type = str::from_utf8(&self.data[self.offset + 4..=self.offset + 7]).unwrap();
        let data = &self.data[self.offset + 8..self.offset + (length as usize)];
        self.offset += length as usize;
        Some(RawField { r#type, data })
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_init() {
        let mut buf = [0u8; 12];
        let (mut p, _) =
            AtemPacket::init(&mut buf, PacketFlag::AckReply.into(), 1234, 5678, None).unwrap();
        p.set_remote_seq_num(9012);
        assert!(p.flags().contains(PacketFlag::AckReply));
        assert_eq!(p.session_id(), 1234);
        assert_eq!(p.local_sequence_number(), 5678);
        assert_eq!(p.remote_sequence_number(), 9012);
        assert_eq!(p.length(), 12);
    }
}
